{+---------------------------------------+}
{|                                       |}
{|      Bugsy Wabbit's Hash List         |}
{|              Freeware                 |}
{|         by Eugene Mayevski            |}
{|                                       |}
{|          http://i.am/Bugsy            |}
{|              Bugsy@i.am               |}
{+---------------------------------------+}

unit BWHashList;  { TBWHashList component. }

interface

uses
  SysUtils,
  MD5Unit;

type EBWHashListError = class (Exception)
     end;

const MaxHashListSize = Maxint div 16;

type
  PHashRecord = ^THashRecord;
  THashRecord = record
{$IFDEF Ver120}
    Hash : array [1..4] of LongWord;
{$ELSE}
    Hash : array [1..4] of integer;
{$ENDIF}
    ItemData : pointer;
  end;

  PHashList = ^THashList;
  THashList = array[0..MaxHashListSize - 1] of PHashRecord;

  TBWHashList = class;

  OnHashDeleteEvent = procedure (Sender: TBWHashList; Data : pointer) of object;

  TBWHashList = class( TObject )
  private
    { Private declarations }
    FList : PHashList;
    FCount : Integer;
    FCapacity : Integer;
    FOnDelete : OnHashDeleteEvent;
    procedure Grow;
    procedure SetCapacity(NewCapacity: Integer);
    function GetItem( Hash : string ) : pointer;
  protected
    { Protected declarations }

  public
    { Public declarations }
    destructor Destroy; override;
    procedure AddItem( Hash : string; Value : Pointer );
    procedure DeleteItem( Hash : string );
    procedure InsertItem( Index : integer; Hash : string; Value : pointer );
    function GetIndex( Hash : string ) : integer;
    procedure Clear;

    property Count : Integer read FCount;  { Public }
    property Capacity : Integer read FCapacity;  { Public }
    function GetByIndex( Index : integer ) : Pointer;
    property Item[Hash : string] : pointer read GetItem;  { Public }
    property OnDelete : OnHashDeleteEvent read FOnDelete write FOnDelete;
  end;  { TBWHashList }

implementation

procedure TBWHashList.AddItem( Hash : string; Value : Pointer );  { public }
begin
  InsertItem(FCount, Hash, Value);
end;  { AddItem }

procedure TBWHashList.DeleteItem( Hash : string );  { public }
var index : integer;
begin
  Index:=GetIndex(Hash);
  if (index<0) or (index>=FCount) then raise EBWHashListError.Create('Hash not found.');
  if assigned(FOnDelete) then FOnDelete(Self, FList^[Index]^.ItemData);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index], (FCount - Index) * SizeOf(Pointer));
end;  { DeleteItem }

procedure TBWHashList.InsertItem( Index : integer; Hash : string; Value : pointer );  { public }
var P : PHashRecord;
    MD5 : TMD5;

begin
  if (Index <0) or (index>FCount) then raise EBWHashListError.Create('Invalid position for HashList.');
  New(P);
  MD5:=TMD5.Create;
  MD5.InputType:=SourceString;
  MD5.InputString:=Hash;
  MD5.pOutputArray:=@P^.Hash;
  MD5.MD5_Hash;
  MD5.Free;
  P^.ItemData:=Value;
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(Pointer));
  FList^[Index] := P;
  Inc(FCount);
end;  { InsertItem }

function TBWHashList.GetItem( Hash : string ) : pointer;  { public }
var i : integer;
begin
  i := GetIndex(Hash);
  if i=-1 then result:=nil else result:=FList^[i]^.ItemData;
end;  { GetItem }

function TBWHashList.GetIndex( Hash : string ) : integer;  { public }
var MD5 : TMD5;
{$IFDEF Ver120}
    arr : array [1..4] of LongWord;
{$ELSE}
    arr : array [1..4] of integer;
{$ENDIF}
    i : integer;
begin
  MD5:=TMD5.Create;
  MD5.InputType:=SourceString;
  MD5.InputString:=Hash;
  MD5.pOutputArray:=@arr;
  MD5.MD5_Hash;
  MD5.Free;
  i:=0;
  while (i<FCount) and (not CompareMem(@Arr,@FList^[i]^.Hash,SizeOf(Integer)*4)) do
        inc(i);
  if i=FCount then result:=-1 else result:=i;
end;  { GetIndex }

procedure TBWHashList.SetCapacity(NewCapacity: Integer);
begin
  if (NewCapacity < FCount) or (NewCapacity > MaxHashListSize) then
    raise EBWHashListError.Create('Invalid BWHashList capacity.');
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
  end;
end;

procedure TBWHashList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

procedure TBWHashList.Clear;  { public }
var J: Integer;
begin
  for j := 0 to FCount-1 do    // Iterate
  begin
    if assigned(FOnDelete) then FOnDelete(Self, FList^[j]^.ItemData);
    Dispose(FList^[j]);
  end;
  FCount:=0;
  SetCapacity(0);
end;  { Clear }

function TBWHashList.GetByIndex( Index : integer ) : Pointer;  { public }
begin
  if (Index<0) or (Index>=FCount)
    then raise EBWHashListError.Create('Invalid index.')
    else result:=FList^[Index]^.ItemData;
end;  { GetByIndex }

destructor TBWHashList.Destroy;
begin
  Clear;
  FreeMem(FList,FCapacity);
  inherited Destroy;
end;  { Destroy }

end.

