{+---------------------------------------+}
{|                                       |}
{|         Bugsy Wabbit's Header         |}
{|              Freeware                 |}
{|         by Eugene Mayevski            |}
{|                                       |}
{|         http://i.am/Bugsy             |}
{|             Bugsy@i.am                |}
{+---------------------------------------+}

unit BWHeader;

interface

{$R BWHeader.res}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, BWTools;

type TBWSectionStyle = (bwhsText, bwhsOwnerDraw);
type TBWSSortMode = (hsmNone, hsmAscend, hsmDescend);
type TBWSAlignment = (hsaLeft, hsaCenter, hsaRight);
type TBWHResizingStates = (trsBegin, trsMove, trsEnd);

type EBWError = class(Exception)
     end;

type TCustomBWHeader = class;
     TBWHeaderSection = class;
     TBWHeaderSectionEvent = procedure (Sender:TCustomBWHeader; Section : TBWHeaderSection) of object;

     TBWHeaderSection = class (TPersistent)
      private
       FVisible : boolean;
       FStyle   : TBWSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TBWSSortMode;
       FAllowClick: boolean;
       FAlignment: TBWSAlignment;
       FText    : string;
       FData    : pointer;
       FOwner   : TCustomBWHeader;
       FTag     : integer;
       procedure SetWidth(value:integer);
       function GetLeft:integer;
       function GetRight:integer;
       procedure SetMaxWidth(value:integer);
       procedure SetMinWidth(value:integer);
       procedure SetText(value:string);
       procedure SetStyle(value:TBWSectionStyle);
       procedure SetSortMode(value:TBWSSortMode);
       procedure SetAlignment(value:TBWSAlignment);
       procedure SetVisible(value:boolean);
       function GetIndex:integer;
      protected
      public
       constructor Create(AOwner:TCustomBWHeader);
       procedure UpdateSection;
       procedure Assign(source:TPersistent); override;
       property Index : integer read GetIndex;
       property Text : string read FText write SetText;
       property Data : pointer read FData write FData;
       property Tag  : integer read FTag write FTag;
       property Style: TBWSectionStyle read FStyle write SetStyle;
       property Left : integer read GetLeft;
       property Width: integer read FWidth write SetWidth;
       property Right: integer read GetRight;
       property MaxWidth: integer read FMaxWidth write SetMaxWidth;
       property MinWidth: integer read FMinWidth write SetMinWidth;
       property SortMode: TBWSSortMode read FSortMode write SetSortMode;
       property AllowClick : boolean read FAllowClick write FAllowClick;
       property Alignment: TBWSAlignment read FAlignment write SetAlignment;
       property Visible : boolean read FVisible write SetVisible;
     end;

    TBWHeaderSections = class (TPersistent)
      private
       FList : TList;
       FOwner: TCustomBWHeader;
       function GetCount:integer;
       function GetSection(index:integer):TBWHeaderSection; virtual;
      protected
       procedure Clear;
       function CreateSection:TBWHeaderSection; virtual;
       procedure WriteData(Stream:TStream); virtual;
       procedure ReadData(Stream:TStream); virtual;
       procedure DefineProperties(Filer: TFiler); override;
       function LastVisibleSection:TBWHeaderSection;
       function GetPrevVisibleSection(Section:TBWHeaderSection):TBWHeaderSection;
      public
       constructor Create(AOwner:TCustomBWHeader);
       destructor Destroy; override;
       procedure Assign(source:TPersistent); override;
       function AddSection:TBWHeaderSection;
       function InsertSection(index:integer):TBWHeaderSection;
       procedure DeleteSection(Section:TBWHeaderSection);
       procedure MoveSection(Section:TBWHeaderSection; NewPos:integer);
       procedure LoadFromStream(Stream:TStream);
       procedure SaveToStream(Stream:TStream);
       procedure SaveToFile(FileName:String);
       procedure LoadFromFile(FileName:String);

       property Item[Index:integer] : TBWHeaderSection read GetSection; default;
       property  Owner : TCustomBWHeader read FOwner;
      published
       property Count : integer read GetCount;
     end;

     TBWSectionRedrawEvent = procedure (Sender:TCustomBWHeader;
                                       Section : TBWHeaderSection; R:TRect; Pressed:boolean) of object;
     TBWSectionResizingEvent = procedure (Sender:TCustomBWHeader;
                                       Section : TBWHeaderSection; State:TBWHResizingStates; Width:integer) of object;
     TBWSectionMoveEvent = procedure (Sender:TCustomBWHeader;
                                      Section : TBWHeaderSection; OldPos, NewPos : integer) of object;

    TCustomBWHeader = class(TCustomControl)
    private
      { Private declarations }
      FSections : TBWHeaderSections;
      FButtons  : boolean;
      FTracking : boolean;
      FAllowDrag   : boolean;
      FPainting    : boolean;
//      FFixClick    : boolean; // Left for future
      FButtonLeft,
      FButtonRight : TBWHeaderSection;
      FAscBmp,
      FDescBmp,
      FLeftBmp,
      FRightBmp : TBitmap;

      FPressCoord  : TPoint;
      FPressed     : boolean;
      FPressedItem : TBWHeaderSection;
      FTrackSection: TBWHeaderSection;
      FResizing    : boolean;
      FResizeSection: TBWHeaderSection;
      FDropSrc,
      FDropTrg     : TBWHeaderSection;
      FLineDC      : HDC;
      FHeaderLine  : integer;
      FLineTab     : integer;
      FResizeOnDrag: boolean;
      FHeaderLineVis: boolean;

      FOnSectionClick  : TBWHeaderSectionEvent;
      FOnSectionResize : TBWHeaderSectionEvent;
      FOnSectionDraw   : TBWSectionRedrawEvent;
      FOnResize        : TNotifyEvent;
      FOnSectionResizing: TBWSectionResizingEvent;
      FOnSectionDelete : TBWHeaderSectionEvent;
      FOnSectionMove   : TBWSectionMoveEvent;
      FOnVisibleChange : TBWHeaderSectionEvent;
      procedure DrawLine;
      procedure AllocateLineDC;
      procedure ReleaseLineDC;
      function GetColumnsWidth: integer;
      procedure InvalidateRight(value:integer);
      procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
      procedure RedrawSection(SectionNum:integer; R:TRect);
      procedure RedrawSections;
      procedure SetSections(value:TBWHeaderSections);
    protected
      { Protected declarations }
      function  CreateSections:TBWHeaderSections; virtual;
      function  InResizeArea (X:integer; var HitSection:TBWHeaderSection):boolean; virtual;
      procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
      procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
      procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
      procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
      procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
      procedure Paint; override;
{$IFDEF VER120}
      procedure Resize; override;
{$ELSE}
      procedure Resize; dynamic;
{$ENDIF}
      procedure DragOver(Source: TObject; X, Y: Integer;
                                 State: TDragState; var Accept: Boolean); override;
      procedure DragDrop(Source: TObject; X, Y: Integer); override;
      procedure DoStartDrag(var DragObject: TDragObject); override;
      procedure DragCanceled; override;

      property SectionsWidth : integer read GetColumnsWidth;
      property Sections : TBWHeaderSections read FSections write SetSections;
      property ResizeOnDrag : boolean read FResizeOnDrag write FResizeOnDrag;
      property Tracking : boolean read FTracking write FTracking;
      property AllowDrag: boolean read FAllowDrag write FAllowDrag;
//      property FixClick : boolean read FFixClick write FFixClick; // Left for future

      property OnResize : TNotifyEvent                   read FOnResize
                                                         write FOnResize;
      property OnSectionShowHide : TBWHeaderSectionEvent read FOnVisibleChange
                                                         write FOnVisibleChange;
      property OnSectionClick : TBWHeaderSectionEvent read FOnSectionClick
                                                      write FOnSectionClick;
      property OnSectionResize : TBWHeaderSectionEvent read FOnSectionResize
                                                      write FOnSectionResize;
      property OnSectionDraw   : TBWSectionRedrawEvent read FOnSectionDraw
                                                      write FOnSectionDraw;
      property OnSectionDelete : TBWHeaderSectionEvent read FOnSectionDelete
                                                       write FOnSectionDelete;
      property OnSectionResizing: TBWSectionResizingEvent read FOnSectionResizing
                                                          write FOnSectionResizing;
      property OnSectionMove : TBWSectionMoveEvent read FOnSectionMove
                                                   write FOnSectionMove;
    public
      { Public declarations }
      constructor Create(AOwner:TComponent); override;
      destructor Destroy; override;
      function GetSectionAt(X,Y:integer):TBWHeaderSection;
      function GetSectionRect(SectionNum:integer):TRect;
      property Canvas;
    published
      { Published declarations }
    end;

    TBWHeader = class (TCustomBWHeader)
      published
        property Color;
        property SectionsWidth;
        property Sections;
        property Tracking;
        property AllowDrag;
        property Align;
        property Enabled;
{$IFDEF VER120}
        property Anchors;
        property Action;
        property ActionLink;
        property Constraints;
        property DockOrientation;
        property Floating;
        property BevelKind;
        property DoubleBuffered;
{$ENDIF}
//        property FixClick; // Left for future
        property Font;
        property ShowHint;
        property ParentFont;
        property ParentShowHint;
        property PopupMenu;
        property ResizeOnDrag;
        property Visible;
        property OnMouseDown;
        property OnMouseMove;
        property OnMouseUp;
        property OnResize;
        property OnSectionClick;
        property OnSectionResize;
        property OnSectionDraw;
        property OnSectionResizing;
        property OnSectionDelete;
        property OnSectionMove;
        property OnSectionShowHide;
    end;

implementation

Const ResizeWidth = 5;

// =============================================================================

constructor TBWHeaderSection.Create;
begin
  inherited Create;
  FStyle:=bwhsText;
  FVisible:=true;
  FOwner:=FOwner;
  FWidth:=80;
  FMinWidth:=0;
  FMaxWidth:=10000;
  FAlignment:=hsaLeft;
  FAllowClick:=true;
end;

procedure TBWHeaderSection.UpdateSection;
var R:TRect;
begin
  if not (FVisible) then exit;
  R:=Rect(Left,0,Left+FWidth,FOwner.Height-1);
  InvalidateRect(FOwner.Handle,@R,false);
end;

procedure TBWHeaderSection.SetSortMode(value:TBWSSortMode);
var i : integer;
begin
  if FSortMode=value then exit;
  for i:=0 to FOwner.FSections.Count-1 do FOwner.FSections[i].FSortMode:=hsmNone;
  FSortMode:=value;
  FOwner.Repaint;
end;

procedure TBWHeaderSection.SetMaxWidth(value:integer);
begin
  if value=FMaxWidth then exit;
  FMaxWidth:=value;
  if FWidth>FMaxWidth then Width:=FMaxWidth;
end;

procedure TBWHeaderSection.SetMinWidth(value:integer);
begin
  if value=FMinWidth then exit;
  FMinWidth:=value;
  if FWidth>FMinWidth then Width:=FMinWidth;
end;

procedure TBWHeaderSection.SetWidth(value:integer);
begin
  if (value>FMaxWidth) or (value<FMinWidth) then exit;
  FWidth:=value;
//  if Self.Index=FOwner.Sections.count-1 then
//  begin
//    r:=Self.Right;
//    if ((r<>FOwner.Width) and (FOwner.ResizeRight)) then FOwner.Width:=r;
//  end;    // Left for ResizeRight Property
  FOwner.InvalidateRight(Self.Index);
end;

procedure TBWHeaderSection.SetVisible;
begin
  FVisible:=value;
  FOwner.InvalidateRight(Self.Index);
  if Assigned (FOwner.FOnVisibleChange) then FOwner.FOnVisibleChange(FOwner, Self); 
end;

function TBWHeaderSection.GetLeft:integer;
var ind:integer;
begin
  ind:=FOwner.Sections.FList.IndexOf(self);
  if ind=0 then result:=0 else result:=FOwner.FSections[ind-1].Right;
end;

function TBWHeaderSection.GetRight:integer;
var i:integer;
    ind:integer;
begin
  ind:=FOwner.FSections.FList.IndexOf(self);
  if ind=0 then i:=0 else i:=FOwner.FSections[ind-1].Right;
  if FVisible then
    result:=i+FWidth else result:=i;
end;

procedure TBWHeaderSection.SetText(value:string);
begin
  if FText=value then exit;
  FText:=value;
  if FStyle=bwhsText then UpdateSection;
end;

procedure TBWHeaderSection.SetStyle(value:TBWSectionStyle);
begin
  if FStyle=value then exit;
  FStyle:=value;
  UpdateSection;
end;

procedure TBWHeaderSection.SetAlignment(value:TBWSAlignment);
begin
  if FAlignment=value then exit;
  FAlignment:=value;
  UpdateSection;
end;

function TBWHeaderSection.GetIndex:integer;
begin
  result:=FOwner.FSections.FList.IndexOf(self);
end;

procedure TBWHeaderSection.Assign(source:TPersistent);
var THS : TBWHeaderSection;
begin
  if Source is TBWHeaderSection then
  begin
    THS:=TBWHeaderSection(source);
    FVisible:=THS.FVisible;
    FStyle:=THS.FStyle;
    FText:=THS.FText;
    FWidth:=THS.FWidth;
    FMinWidth:=THS.FMinWidth;
    FMaxWidth:=THS.FMaxWidth;
    FSortMode:=THS.FSortMode;
    FAllowClick:=THS.FAllowClick;
    FAlignment:=THS.FAlignment;
  end else inherited;
end;

// =============================================================================

constructor TBWHeaderSections.Create(AOwner:TCustomBWHeader);
begin
  inherited Create;
  FOwner := AOwner;
  FList:= TList.Create;
end;

destructor TBWHeaderSections.Destroy;
begin
  Clear;
  FList.Free;
  inherited destroy;
end;

procedure TBWHeaderSections.Clear;
begin
  while FList.Count>0 do
  begin
    TBWHeaderSection(FList[0]).Free;
    FList.Delete(0);
  end;
end;

procedure TBWHeaderSections.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, Count>0);
end;

function TBWHeaderSections.CreateSection:TBWHeaderSection;
begin
  result:=TBWHeaderSection.Create(FOwner);
end;

function TBWHeaderSections.AddSection:TBWHeaderSection;
begin
  result:=InsertSection(FList.Count);
end;

function TBWHeaderSections.InsertSection(index:integer):TBWHeaderSection;
begin
  result:=CreateSection;
  Result.FOwner:=FOwner;
  FList.Insert(index, result);
  FOwner.InvalidateRight(index);
  FOwner.Update;
end;

procedure TBWHeaderSections.DeleteSection(Section:TBWHeaderSection);
begin
 if Section=nil then Raise EBWError.Create('Invalid section');
 FList.Remove(Section);
 if assigned(FOwner.FOnSectionDelete) then FOwner.FOnSectionDelete(FOwner,Section);
 Section.Free;
 FOwner.Invalidate;
 FOwner.Update;
end;

procedure TBWHeaderSections.MoveSection(Section:TBWHeaderSection; NewPos:integer);
var i:integer;
begin
  i:=FList.IndexOf(Section);
  FList.Move(i,NewPos);
  if assigned(FOwner.FOnSectionMove) then FOwner.FOnSectionMove(FOwner, TBWHeaderSection(FList[i]), i, NewPos);
  FOwner.InvalidateRight(min(i,NewPos));
  FOwner.Update;
end;

function TBWHeaderSections.GetCount:integer;
begin
  result:=FList.Count;
end;

function TBWHeaderSections.GetSection;
begin
  result:=TBWHeaderSection(FList[index]);
end;

type PSectionData = ^TSectionData;
     TSectionData = record
       FVisible : boolean;
       FStyle   : TBWSectionStyle;
       FWidth   : integer;
       FMinWidth,
       FMaxWidth: integer;
       FSortMode: TBWSSortMode;
       FAllowClick: boolean;
       FAlignment: TBWSAlignment;
     end;

procedure TBWHeaderSections.WriteData(Stream:TStream);
var i,j:integer;
    P:PSectionData;
    P1:PChar;
    S:String;
begin
  inherited;
  GetMem(P,SizeOf(TSectionData));
  i:=Count;
  Stream.WriteBuffer(i,SizeOf(integer));
  for i:=0 to count-1 do
  begin
    MoveMemory(P,@(TBWHeaderSection(FList[i]).FVisible),SizeOf(TSectionData));
    Stream.WriteBuffer(p^,SizeOf(TSectionData));
    if P^.FStyle=bwhsText then
    begin
      S:=TBWHeaderSection(FList[i]).Text;
      j:=Length(S)+1;
      GetMem(P1,j);
      StrPCopy(P1,S);
      Stream.WriteBuffer(j,SizeOf(integer));
      Stream.WriteBuffer(P1^,j);
      FreeMem(P1,j);
    end;
  end;
  FreeMem(P,SizeOf(TSectionData));
end;

procedure TBWHeaderSections.ReadData(Stream:TStream);
var i,j:integer;
    P:PSectionData;
    P1:PChar;
    S:String;
    THS:TBWHeaderSection;
begin
  inherited;
  Clear;
  Stream.ReadBuffer(i,SizeOf(integer));
  GetMem(P,SizeOf(TSectionData));
  while i>0 do
  begin
    Stream.ReadBuffer(p^,SizeOf(TSectionData));
    THS:=AddSection;
    MoveMemory(@THS.FVisible,pointer(P),SizeOf(TSectionData));
    if P^.FStyle=bwhsText then
    begin
      Stream.ReadBuffer(J,SizeOf(integer));
      GetMem(P1,j);
      Stream.ReadBuffer(P1^,j);
      S:=StrPas(P1);
      THS.FText:=S;
      FreeMem(P1,j);
    end;
    dec(i);
  end;
  FreeMem(P,SizeOf(TSectionData));
  FOwner.Update;
end;

procedure TBWHeaderSections.LoadFromStream(Stream:TStream);
begin
  ReadData(Stream);
  if FOwner<>nil then FOwner.Invalidate;    
end;

procedure TBWHeaderSections.SaveToStream(Stream:TStream);
begin
  WriteData(Stream);
end;

procedure TBWHeaderSections.SaveToFile(FileName:String);
var T:TStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);
    SaveToStream(T);
  finally
    T.Free;
  end;
end;

procedure TBWHeaderSections.LoadFromFile(FileName:String);
var T:TStream;
begin
  T:=nil;
  try
    T:=TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    LoadFromStream(T);
  finally
    T.Free;
  end;
end;

function TBWHeaderSections.LastVisibleSection:TBWHeaderSection;
begin
  if Item[Count-1].Visible then result:=Item[Count-1] else
     result:=GetPrevVisibleSection(Item[Count-1]);
end;

function TBWHeaderSections.GetPrevVisibleSection;
var I:Integer;
begin
  result:=nil;
  I:=FList.IndexOf(Section)-1;
  if i<0 then exit;
  while i>=0 do
  begin
    if Item[i].Visible then
    begin
      result:=Item[i];
      break;
    end;
    dec(i);
  end;    // while
end;

procedure TBWHeaderSections.Assign;
var IT: TBWHeaderSections;
    j : integer;
begin
  if source is TBWHeaderSections then
  begin
    IT:=TBWHeaderSections(source);
    // delete old sections
    for j := 0 to FList.Count-1 do DeleteSection(Item[j]);
    // add new sections
    for j := 0 to IT.Count-1 do CreateSection.Assign(IT.Item[j]);
  end else inherited;
end;

// =============================================================================

constructor TCustomBWHeader.Create(AOwner:TComponent);
begin
  inherited;
  ControlStyle:=[csClickEvents, csDoubleClicks, csCaptureMouse,
   csDesignInteractive, csReplicatable, csReflector];
  FSections:=CreateSections;
  TabStop:=false;
//  Align:=alTop;
  Height:=abs(Font.Height)+5;
  DragMode:=dmManual;
  FPressed:=false;
  DragCursor:=Cursor;
  FAscBmp:=TBitmap.Create;
  FDescBmp:=TBitmap.Create;
  FLeftBmp:=TBitmap.Create;
  FRightBmp:=TBitmap.Create;
  FPressed:=false;
  FTracking:=true;
  FAllowDrag:=true;
  FResizeOnDrag:=false;
//  FFixClick:=false; // Left for future
  FAscBmp.LoadFromResourceName(HInstance,'ASCBMP');
  FDescBmp.LoadFromResourceName(HInstance,'DESCBMP');
  FLeftBmp.LoadFromResourceName(HInstance,'LEFTBMP');
  FRightBmp.LoadFromResourceName(HInstance,'RIGHTBMP');
end;

destructor TCustomBWHeader.Destroy;
begin
  FSections.Free;
  FAscBmp.Free;
  FDescBmp.Free;
  FLeftBmp.Free;
  FRightBmp.Free;
  inherited Destroy;
end;

function TCustomBWHeader.CreateSections;
begin
  result:=TBWHeaderSections.Create(self);
end;

function TCustomBWHeader.GetSectionAt(X,Y:integer):TBWHeaderSection;
var i:integer;
    j:integer;
begin
  j:=0;
  result:=nil;
  for i:=0 to FSections.FList.Count-1 do
  begin
    if FSections[i].Visible then j:=j+FSections[i].Width;
    if j>X then
    begin
      result:=FSections[i];
      exit;
    end;
  end;
end;

procedure TCustomBWHeader.InvalidateRight(value:integer);
var R:TRect;
begin
  if not inRange(0,Sections.Count-1,value) then Raise EBWError.Create('Index out of bounds');
  R:=Rect(Sections[value].Left, 0, ClientWidth, ClientHeight);
  InvalidateRect(handle,@R,false);
  Update;
end;

function TCustomBWHeader.GetColumnsWidth: integer;
var i:integer;
begin
  result:=0;
  i:=FSections.Count-1;
  while i>=0 do
  begin
    if FSections[i].FVisible then
    begin
      result:=FSections[i].Right;
      exit;
    end;
    dec(i);
  end;
end;

function TCustomBWHeader.InResizeArea(X:integer; var HitSection:TBWHeaderSection):boolean;
var THS : TBWHeaderSection;
    l : integer;
begin
  THS:=GetSectionAt(X,0);
  HitSection:=THS;
  if THS=nil then
  begin
    if Sections.Count>0 then THS:=Sections.LastVisibleSection;
    if THS<>nil then
    begin
      l:=THS.Right;
      if L+ResizeWidth>X then
      begin
        result:=true;
        HitSection:=THS;
        exit;
      end;
    end;
    result:=false;
    HitSection:=nil;
    exit;
  end
  else
  begin
    if THS.Width<ResizeWidth*3 then result:=false else
    begin
      l:=THS.Left;
      if L+ResizeWidth>X then
      begin
        if Sections.GetPrevVisibleSection(THS)=nil then
        begin
          result:=false;
          HitSection:=THS;
          exit;
        end;
        result:=true;
        HitSection:=Sections.GetPrevVisibleSection(THS);
        exit;
      end;
      if L+THS.FWidth-ResizeWidth<=X then
      begin
        result:=true;
        exit;
      end;
      result:=false;
    end;
  end;
end;

procedure TCustomBWHeader.WMLButtonDown(var Message: TWMLButtonDown);
var THS,THS1 : TBWHeaderSection;
begin
  if not (csNoStdEvents in ControlStyle) then
  begin
    if InResizeArea(Message.XPos, THS) then
    begin
      FResizing:=true;
      FResizeSection:=THS;
      THS1:=FTrackSection;
      FTrackSection:=nil;
      if THS1<>nil then THS1.UpdateSection;
      Screen.Cursor:=crHSplit;
      SetCapture(Handle);
      FHeaderLine:=FResizeSection.Right;
      FLineTab:=Message.XPos-FHeaderLine;
      if (not ResizeOnDrag) then
      begin
        AllocateLineDC;
        DrawLine;
      end;
      if Assigned(FOnSectionResizing)
         then FOnSectionResizing(Self, THS, trsBegin, THS.Width);
    end // Resizing
    else
    begin
      FPressCoord:=Point(Message.XPos,Message.YPos);
      FPressed:=true;
      if FPressedItem<>nil then
      begin
        THS:=FPressedItem;
        FPressedItem:=nil;
        THS.UpdateSection;
      end;
      FPressedItem:=GetSectionAt(Message.XPos, Message.YPos);
      if (FPressedItem<>nil) and (FPressedItem.FAllowClick or FPressed)
         then FPressedItem.UpdateSection
         else FPressedItem:=nil;
    end; // not Resizing
  end;
  inherited;
end;

procedure TCustomBWHeader.WMMouseMove(var Message: TWMMouseMove);
var THS : TBWHeaderSection;
    NewWidth,L : integer;
begin
  if (FPressed) and (FAllowDrag) then
  begin
    if (FPressedItem<>nil)
       and (sqrt(sqr(FPressCoord.X-Message.XPos)+sqr(FPressCoord.Y-Message.YPos))>=5) then
    begin
      FDropSrc:=FPressedItem;
      FPressed:=false;
      BeginDrag(true);
    end;
  end;
  if (FTracking) and (not FResizing) and (not (csNoStdEvents in ControlStyle)) then
  begin
    THS:=FTrackSection;
    FTrackSection:=GetSectionAt(Message.XPos, Message.YPos);
    if THS<>FTrackSection then
    begin
      if THS<>nil then THS.UpdateSection;
      if FTrackSection<>nil then FTrackSection.UpdateSection;
      Update;
    end;
  end;
  if (not FPressed) and (not FResizing) then
  begin
    if InResizeArea(Message.XPos,THS) then
    begin
      if Screen.Cursor<>crHSplit then Screen.Cursor:=crHSplit;
    end
    else if Screen.Cursor=crHSplit then Screen.Cursor:=Cursor;
  end;
  if FResizing then
  begin
    L:=FResizeSection.Left;
    if (Message.XPos-FLineTab>=L) { and ((Message.XPos<Width) or ResizeRight) }then
    begin
      NewWidth:=Message.XPos-L-FLineTab;
      with FResizeSection do
        if InRange(MinWidth,MaxWidth,NewWidth) then
        begin
          if FResizeOnDrag then
          begin
            Width:=NewWidth;
            InvalidateRight(FResizeSection.Index);
            Update;
            if Assigned(FOnSectionResizing)
               then FOnSectionResizing(Self, FResizeSection, trsMove, NewWidth);
          end
          else
          begin
            DrawLine;
            FHeaderLine:=L+NewWidth;
            DrawLine;
            if Assigned(FOnSectionResizing)
               then FOnSectionResizing(Self, FResizeSection, trsMove, NewWidth);
          end;
        end;
    end; // really in client width
  end;
  inherited;
end;

procedure TCustomBWHeader.WMLButtonUp(var Message: TWMLButtonUp);
var THS : TBWHeaderSection;
begin
  if not (csNoStdEvents in ControlStyle) then
  begin
    if FResizing then
    begin
      FResizing:=false;
      Screen.Cursor:=Cursor;
      ReleaseCapture;
      if (not FResizeOnDrag) then
      begin
        if (FHeaderLineVis) then DrawLine;
        ReleaseLineDC;
        FResizeSection.Width:=FHeaderLine-FResizeSection.Left;
      end;
      if Assigned(FOnSectionResizing)
         then FOnSectionResizing(Self, FResizeSection, trsEnd, FResizeSection.Width);
      if Assigned(FOnSectionResize) then FOnSectionResize(Self, FResizeSection);
      FResizeSection:=nil;
      Invalidate;
      Update;
    end;
    if FDropSrc=nil then
    begin
      FPressed:=false;
      THS:=FPressedItem;
//      if not FFixClick then
      begin
        FPressedItem:=nil;
        if THS<>nil then THS.UpdateSection;
      end;
      if THS<>nil then
         if (GetSectionAt(FPressCoord.X,FPressCoord.Y)=
             GetSectionAt(Message.XPos,Message.YPos))
             and (Assigned(FOnSectionClick))
             then FOnSectionClick(Self, THS);
    end;
  end;
  inherited;
end;

procedure TCustomBWHeader.DoStartDrag(var DragObject: TDragObject);
begin
  FDropTrg := nil;
  inherited DoStartDrag(DragObject);
end;

procedure TCustomBWHeader.DragCanceled;
var THS: TBWHeaderSection;
begin
  inherited;
  FDropTrg:=nil;
  FDropSrc:=nil;
  THS:=FPressedItem;
  FPressedItem:=nil;
  if THS<>nil then THS.UpdateSection;
end;

procedure TCustomBWHeader.DragDrop;
var
  THS : TBWHeaderSection;
  NewPos : integer;

begin
  THS:=FPressedItem;
  FPressedItem:=nil;
  if THS<>nil then THS.UpdateSection;
  FDropTrg:=GetSectionAt(X,Y);
  if (FDropSrc=FDropTrg) or (FDropSrc=nil) then exit;
  if FDropTrg=nil then NewPos:=Sections.Count-1 else NewPos:=FSections.FList.IndexOf(FDropTrg);
  FSections.MoveSection(FDropSrc, NewPos);
  FDropSrc:=nil;
  FDropTrg:=nil;
  Update;
  inherited;
end;

procedure TCustomBWHeader.CMMouseLeave;
var THS:TBWHeaderSection;
begin
  FDropTrg:=nil;
  FPressed:=false;
  if FPressedItem<>nil then
  begin
    THS:=FPressedItem;
    THS.UpdateSection;
  end;
  if FTrackSection<>nil then
  begin
    THS:=FTrackSection;
    FTrackSection:=nil;
    THS.UpdateSection;
  end;
  if (not FResizing) and (Screen.Cursor=crHSplit) then Screen.Cursor:=Cursor;    
end;

procedure TCustomBWHeader.CMMouseEnter;
begin
  inherited;
  if FPressedItem<>nil then
  begin
    FPressed:=true;
    FPressedItem.UpdateSection;
  end;  
end;

procedure TCustomBWHeader.CMDrag(var Message: TCMDrag);
var TSI : TBWHeaderSection;
begin
  inherited;
  with Message, DragRec^ do
    case DragMessage of
//      dmDragMove: with ScreenToClient(Pos) do DoDragOver(Source, X, Y, Message.Result<>0);
      dmDragLeave:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
        end;
      dmDragDrop:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
        end;
      dmDragCancel:
        begin
          TSI:=FDropTrg;
          FDropTrg := nil;
          if TSI<>nil then
          begin
            TSI.UpdateSection;
            Update;
          end;
          FDropSrc:=nil;
        end;
    end;
end;

function TCustomBWHeader.GetSectionRect(SectionNum:integer):TRect;
var THS:TBWHeaderSection;
    i:integer;
begin
  THS:=TBWHeaderSection(Sections[SectionNum]);
  i:=THS.Left;
  result:=Rect(i,0,i+THS.Width,ClientHeight);
end;

procedure TCustomBWHeader.RedrawSection(SectionNum:integer; R:TRect);
var TS:TBWHeaderSection;
    R1,R2:TRect;
    w:integer;
    DC:THandle;
    s:string;
    SaveCol : TColor;
    SaveStyle : TFontStyles;
    AL: integer;

begin
  TS:=Sections[SectionNum];
  Canvas.Brush.Color := Color;
  Canvas.FillRect(R);
  Canvas.Brush.Color := clBtnFace;
  Canvas.Brush.Style := bsSolid;
  DC := Canvas.Handle;
  if (TS=FPressedItem) and (FPressed=true) then
  begin
    DrawEdge(DC, R, BDR_SUNKENINNER, BF_TOPLEFT);              { black     }
    DrawEdge(DC, R, BDR_SUNKENOUTER, BF_BOTTOMRIGHT);          { btnhilite }
  end
  else
  begin
    DrawEdge(DC, R, BDR_RAISEDOUTER, BF_BOTTOMRIGHT);          { black }
    Dec(R.Bottom);
    Dec(R.Right);
    DrawEdge(DC, R, BDR_RAISEDINNER, BF_TOPLEFT);              { btnhilite }
    Inc(R.Top);
    Inc(R.Left);
    DrawEdge(DC, R, BDR_RAISEDINNER, BF_BOTTOMRIGHT or BF_MIDDLE); { btnshadow }
  end;
  inc(R.Top); inc(R.Left); dec(R.Bottom); dec(R.Right,4);
  R1:=R;
  if (TS=FPressedItem) and (FPressed=true) then inc(R1.Left,2);
  if TS.FSortMode<>hsmNone then
  begin
    if R1.Right-8<R1.Left+ResizeWidth then
    begin
      w:=0;
      R1.Right:=R1.Left+ResizeWidth;
    end else
    begin
      w:=8;
      dec(R1.Right,9);
    end;
    R2:=Rect(R1.Right, (R1.Bottom div 2)-2, R1.Right+W, (R1.Bottom div 2)+3);
    Canvas.Brush.Color:=Color;
    if w>0 then
    begin
      if TS.FSortMode=hsmAscend then
         Canvas.BrushCopy(R2,FAscBmp,Rect(0,0,8,5),FAscBmp.TransparentColor)
      else
        Canvas.BrushCopy(R2,FDescBmp,Rect(0,0,8,5),FDescBmp.TransparentColor);
    end;    
  end else dec(R1.Right,ResizeWidth);
  inc(R1.Left,ResizeWidth);
  if R1.Right<R1.Left then exit;
  if TS.FStyle=bwhsOwnerDraw then
  begin
    if Assigned(FOnSectionDraw) then FOnSectionDraw(Self,TS,R1,(TS=FPressedItem) and (FPressed=true));
  end
  else
  begin
    s:=TS.Text;
    Al:=DT_LEFT;
    case TS.FAlignment of
      hsaCenter: Al:=DT_CENTER;
      hsaRight:  Al:=DT_RIGHT;
    end;    // case
    if TS=FTrackSection then
    begin
      SaveCol:=Canvas.Font.Color;
      SaveStyle:=Canvas.Font.Style;
      Canvas.Font.Color:=clBlue;
      Canvas.Font.Style:=Canvas.Font.Style+[fsUnderline];
      DrawTextEx(Canvas.Handle, PChar(S), -1, R1,
            DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_END_ELLIPSIS or AL, nil);
      Canvas.Font.Color:=SaveCol;
      Canvas.Font.Style:=SaveStyle;
    end else
      DrawTextEx(Canvas.Handle, PChar(S), -1, R1,
            DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER or DT_END_ELLIPSIS or AL, nil);
  end;
end;

procedure TCustomBWHeader.RedrawSections;
var R:TRect;
    R1:TRect;
    i:integer;

begin
  R:=BoundsRect;
  Canvas.Font.Assign(Font);
  for i:=0 to Sections.Count-1 do
  begin
    R:=GetSectionRect(i);
    //if IntersectRect(R1, R, Canvas.ClipRect) then
    RedrawSection(I,R);
  end;
  R1:=R;
  R1.Left:=SectionsWidth;
  R1.Right:=Width;
  Canvas.Brush.Color:=Color;
  Canvas.FillRect(R1);
  DrawEdge(Canvas.Handle, R1, BDR_RAISEDINNER, BF_RECT{BF_BOTTOMRIGHT});
end;

procedure TCustomBWHeader.Resize;
begin
  if Assigned(FOnResize) then FOnResize(Self);
end;

procedure TCustomBWHeader.DragOver;
begin
  inherited;
//  if not assigned(OnDragOver) then
    Accept:= (Source=(self as TCustomBWHeader));
  Accept:=true;
end;

procedure TCustomBWHeader.AllocateLineDC;
begin
  FLineDC := GetDCEx(Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS
    or DCX_LOCKWINDOWUPDATE);
end;

procedure TCustomBWHeader.ReleaseLineDC;
begin
  ReleaseDC(Handle, FLineDC);
end;

procedure TCustomBWHeader.DrawLine;
begin
  FHeaderLineVis:= not FHeaderLineVis;
  PatBlt(FLineDC, FHeaderLine, 0, 1, Height, PATINVERT);
end;

procedure TCustomBWHeader.Paint;
begin
  inherited;
  FPainting:=true;
  RedrawSections;
  FPainting:=false;
end;

procedure TCustomBWHeader.SetSections;
begin
  FSections.Assign(value);
end;

end.
